@@ -128,13 +128,8 @@ def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
     :cfgfile: config file in the format of dictConfig.
     :return: logger object.
     """
-    try:
-        os.makedirs(logdir)
-    except:
-        pass
-
+    os.makedirs(logdir, exist_ok=True)
     cfg_file = cfgfile or join(_script_dir, 'logging.json')
-    logging_config = None
     try:
         if sys.version_info.major > 2:
             with open(cfg_file, 'r', encoding=TXT_CODEC,
@@ -144,8 +139,7 @@ def build_default_logger(logdir, name=None, cfgfile=None, verbose=False):
             with open(cfg_file, 'rU') as f:
                 text = f.read()
         # Add object_pairs_hook=coll.OrderedDict hook for py3.5 and lower.
-        logging_config = json.loads(text,
-                                    object_pairs_hook=collections.OrderedDict)
+        logging_config = json.loads(text, object_pairs_hook=collections.OrderedDict)
         logging_config['handlers']['file']['filename'] = join(logdir, logging_config['handlers']['file']['filename'])
     except Exception:
         filename = name or basename(basename(logdir.strip('\\/')))
@@ -314,6 +308,7 @@ def load_json(path):
     # Add object_pairs_hook=collections.OrderedDict hook for py3.5 and lower.
     return json.loads(text, object_pairs_hook=collections.OrderedDict)
 
+
 def load_json_obj(path):
     """
     Load Json configuration file.
@@ -444,7 +439,7 @@ def query_yes_no(question, default=True):
     Side Effects:
         Blocks program execution until valid input(y/n) is given.
     """
-    input_ = input if sys.version_info.major > 2 else raw_input
+    input_ = input
     yes_list = ['yes', 'y']
     no_list = ['no', 'n']
 
@@ -535,17 +530,17 @@ def logprogress(msg='Task', loghook=_logger.info, errorhook=_logger.error):
     return wrap
 
 
-def logcall(msg='trace', loghook=_logger.debug, errorhook=_logger.error):
+def logcall(msg='trace', logger=_logger):
     def wrap(function):
         @functools.wraps(function)
         def wrapper(*args, **kwargs):
-            loghook("Calling function '{}' with args={} kwargs={}: {}.".format(function.__name__, args, kwargs, msg))
+            logger.debug("Calling function '{}' with args={} kwargs={}: {}.".format(function.__name__, args, kwargs, msg))
             try:
                 response = function(*args, **kwargs)
             except Exception as error:
-                loghook("Function '{}' raised {} with error '{}'.".format(function.__name__, error.__class__.__name__, str(error)))
+                logger.error("Function '{}' raised {} with error '{}'.".format(function.__name__, error.__class__.__name__, str(error)))
                 raise error
-            loghook("Function '{}' returned {}.".format(function.__name__, response))
+            logger.debug("Function '{}' returned {}.".format(function.__name__, response))
             return response
         return wrapper
     return wrap
@@ -651,10 +646,7 @@ def execute_concurrency(worker, shared, lock, algorithm):
 def profile_runs(funcname, modulefile, nruns=5):
     module_name = splitext(basename(modulefile))[0]
     stats_dir = join(abspath(dirname(modulefile)), 'stats')
-    try:
-        os.makedirs(stats_dir)
-    except Exception:
-        traceback.print_exc()
+    os.makedirs(stats_dir, exist_ok=True)
     for i in range(nruns):
         stats_file = join(stats_dir, 'profile_{}_{}.pstats.log'.format(funcname, i))
         profile.runctx('import {}; print({}, {}.{}())'.format(module_name, i, module_name, funcname), globals(), locals(), stats_file)
@@ -678,12 +670,10 @@ def get_local_tmp_dir():
         return join(expanduser('~'), 'Library', 'Caches')
     elif plat == 'Linux':
         return '/tmp'
-    else:
-        raise NotImplementedError(f'unsupported platform: {plat}')
+    raise NotImplementedError(f'unsupported platform: {plat}')
 
 
 def write_plist_fields(cfg_file, my_map):
-    plist = {}
     with open(cfg_file, 'rb') as fp:
         plist = plistlib.load(fp, fmt=plistlib.FMT_XML)
     plist.update(my_map)
@@ -692,7 +682,6 @@ def write_plist_fields(cfg_file, my_map):
 
 
 def substitute_keywords_in_file(file, str_map, useliteral=False):
-    updated = ''
     with open(file) as f:
         original = f.read()
         if not useliteral:
@@ -700,7 +689,6 @@ def substitute_keywords_in_file(file, str_map, useliteral=False):
         else:
             updated = original
             for src, dest in str_map.items():
-                tmp = updated
                 updated = updated.replace(src, dest)
     with open(file, 'w') as f:
         f.write(updated)
@@ -708,10 +696,10 @@ def substitute_keywords_in_file(file, str_map, useliteral=False):
 
 def is_uuid(text, version=4):
     try:
-        uuid_obj = uuid.UUID(text, version=4)
-        return True
+        _ = uuid.UUID(text, version=version)
     except ValueError:
         return False
+    return True
 
 
 def get_clipboard_content():
@@ -751,10 +739,7 @@ def convert_from_wine_path(path):
 
 
 def kill_process_by_name(name):
-    if platform.system() == 'Windows':
-        cmd = ['taskkill', '/IM', name, '/F']
-    else:
-        cmd = ['pkill', name]
+    cmd = ['taskkill', '/IM', name, '/F'] if platform.system() == 'Windows' else ['pkill', name]
     subprocess.run(cmd)
 
 
@@ -763,7 +748,6 @@ def init_translator(localedir, domain='all', langs=None):
     - select locale and set up translator based on system language
     - the leading language in langs, if any, is selected to override current locale
     """
-    cur_langs = None
     if langs:
         cur_langs = langs
     else:
@@ -794,45 +778,43 @@ def match_files_except_lines(file1, file2, excluded=None):
 
 class RerunLock:
     """Lock process from reentering when seeing lock file on disk."""
-    def __init__(self, name, folder=None, infohook=_logger.info, warnhook=_logger.warning, errorhook=_logger.error):
+    def __init__(self, name, folder=None, logger=_logger):
         os.makedirs(folder, exist_ok=True)
-        filename = 'lock_{}'.format(name) if name else 'lock_{}'.format(next(tempfile._get_candidate_names()))
-        self.lockFile = join(folder, filename) if folder else join(_script_dir, os.pardir, filename)
-        self.infoHook = infohook
-        self.warnHook = warnhook
-        self.errorHook = errorhook
+        filename = f'lock_{name}.json' if name else 'lock_{}.json'.format(next(tempfile._get_candidate_names()))
+        self.lockFile = osp.join(folder, filename) if folder else join(get_local_tmp_dir(), filename)
+        self.infoHook = logger.info
+        self.warnHook = logger.warning
+        self.errorHook = logger.error
 
     def lock(self):
         if not self.is_locked():
-            with open(self.lockFile, 'w') as f:
-                pass
+            save_json(self.lockFile, {'pid': os.getpid()})
             return True
         else:
-            self.warnHook('Will not run the script while it is active with pid: {}.'.format(os.getpid()))
+            self.warnHook('Locked by pid: {}. Will stay locked until it ends.'.format(os.getpid()))
             return False
 
     def unlock(self):
         try:
             os.remove(self.lockFile)
         except FileNotFoundError:
-            self.warnHook('Script reentrance is already enabled.')
+            self.warnHook('Already unlocked. Aborted.')
         except Exception:
             failure = traceback.format_exc()
-            self.errorHook('{}\nFailed to unlock the script. You must delete the lock by hand: {}.'.format(failure, self.lockFile))
+            self.errorHook('{}\nFailed to unlock. Must delete the lock by hand: {}.'.format(failure, self.lockFile))
 
     def is_locked(self):
-        return exists(self.lockFile)
+        return osp.isfile(self.lockFile)
 
 
-def rerun_lock(name, folder=None, infohook=_logger.info, warnhook=_logger.warning, errorhook=_logger.error):
+def rerun_lock(name, folder=None, logger=_logger):
     """Decorator for reentrance locking on functions"""
     def decorator(f):
         @functools.wraps(f)
         def wrapper(*args, **kwargs):
             my_lock = None
-            ret = None
             try:
-                my_lock = RerunLock(name, folder, infohook, warnhook, errorhook)
+                my_lock = RerunLock(name, folder, logger)
                 if not my_lock.lock():
                     return 1
                 ret = f(*args, **kwargs)
@@ -846,90 +828,105 @@ def rerun_lock(name, folder=None, infohook=_logger.info, warnhook=_logger.warnin
     return decorator
 
 
-def append_to_os_paths(bindir):
+def append_to_os_paths(bindir, usesyspath=True):
     """
     On macOS, PATH update will only take effect after calling `source ~/.bash_profile` directly in shell. It won't work 
     """
     if platform.system() == 'Windows':
         import winreg
-        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
-            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
-                user_paths, _ = winreg.QueryValueEx(key, 'Path')
-                if bindir not in user_paths:
-                    if user_paths[-1] != ';':
-                        user_paths += ';'
-                    user_paths += f'{bindir}'
-                    winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+        root_key = winreg.HKEY_LOCAL_MACHINE if usesyspath else winreg.HKEY_CURRENT_USER
+        sub_key = r'SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment' if usesyspath else r'Environment'
+        path_var = 'Path'
+        separator = ';'
+        with winreg.ConnectRegistry(None, root_key):
+            with winreg.OpenKey(root_key, sub_key, 0, winreg.KEY_ALL_ACCESS) as key:
+                env_paths, _ = winreg.QueryValueEx(key, path_var)
+                matches = [path for path in env_paths.split(separator) if path.lower() == bindir.lower()]
+                if matches:
+                    return
+                if env_paths[-1] != separator:
+                    env_paths += separator
+                env_paths += f'{bindir}{separator}'
+                winreg.SetValueEx(key, path_var, 0, winreg.REG_EXPAND_SZ, env_paths)
     else:
+        path_var = 'PATH'
         cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
-        if bindir in os.environ['PATH']:
+        if bindir in os.environ[path_var]:
             return
         with open(cfg_file, 'a') as fp:
-            fp.write(f'\nexport PATH="$PATH:{bindir}"\n\n')
-    os.environ['PATH'] += os.pathsep + bindir
+            fp.write(f'\nexport {path_var}="${path_var}:{bindir}"\n\n')
+    os.environ[path_var] += os.pathsep + bindir
 
 
-def prepend_to_os_paths(bindir):
+def prepend_to_os_paths(bindir, usesyspath=True):
     if platform.system() == 'Windows':
         import winreg
-        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
-            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
-                user_paths, _ = winreg.QueryValueEx(key, 'Path')
-                if bindir not in user_paths:
-                    user_paths = f'{bindir};' + user_paths
-                    winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+        root_key = winreg.HKEY_LOCAL_MACHINE if usesyspath else winreg.HKEY_CURRENT_USER
+        sub_key = r'SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment' if usesyspath else r'Environment'
+        path_var = 'Path'
+        separator = ';'
+        with winreg.ConnectRegistry(None, root_key):
+            with winreg.OpenKey(root_key, sub_key, 0, winreg.KEY_ALL_ACCESS) as key:
+                env_paths, _ = winreg.QueryValueEx(key, path_var)
+                matches = [path for path in env_paths.split(separator) if path.lower() == bindir.lower()]
+                if matches:
+                    return
+                env_paths = f'{bindir}{separator}' + env_paths
+                winreg.SetValueEx(key, path_var, 0, winreg.REG_EXPAND_SZ, env_paths)
     else:
+        path_var = 'PATH'
         cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
-        if bindir in os.environ['PATH']:
+        if bindir in os.environ[path_var]:
             return
-        lines = []
         with open(cfg_file) as fp:
             lines = fp.readlines()
-        lines = [f'\nexport PATH="{bindir}:$PATH"\n\n'] + lines
+        lines = [f'\nexport {path_var}="{bindir}:${path_var}"\n\n'] + lines
         with open(cfg_file, 'w') as fp:
             fp.writelines(lines)
-    os.environ['PATH'] = bindir + os.pathsep + os.environ['PATH']
+    os.environ[path_var] = bindir + os.pathsep + os.environ[path_var]
 
 
-def remove_from_os_paths(bindir):
+def remove_from_os_paths(bindir, usesyspath=True):
     if platform.system() == 'Windows':
         import winreg
-        with winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER) as reg:
-            with winreg.OpenKey(winreg.HKEY_CURRENT_USER, r'Environment', 0, winreg.KEY_ALL_ACCESS) as key:
-                user_paths, _ = winreg.QueryValueEx(key, 'Path')
-                start = user_paths.find(bindir)
-                if not bindir in user_paths:
+        root_key = winreg.HKEY_LOCAL_MACHINE if usesyspath else winreg.HKEY_CURRENT_USER
+        sub_key = r'SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment' if usesyspath else r'Environment'
+        separator = ';'
+        path_var = 'Path'
+        with winreg.ConnectRegistry(None, root_key):
+            with winreg.OpenKey(root_key, sub_key, 0, winreg.KEY_ALL_ACCESS) as key:
+                env_paths, _ = winreg.QueryValueEx(key, path_var)
+                matches = [path for path in env_paths.split(separator) if path.lower() == bindir.lower()]
+                if not matches:
                     return
-                import re
-                # escape to handle metachars
-                pattern = re.compile('|'.join(map(re.escape, bindir)))
-                pattern.sub('', user_paths)
-                winreg.SetValueEx(key, 'Path', 0, winreg.REG_EXPAND_SZ, user_paths)
+                keepers = [path for path in env_paths.split(separator) if path.lower() != bindir.lower()]
+                env_paths = separator.join(keepers)
+                winreg.SetValueEx(key, path_var, 0, winreg.REG_EXPAND_SZ, env_paths)
     else:
+        path_var = 'PATH'
         cfg_file = os.path.expanduser('~/.bash_profile') if platform.system() == 'Darwin' else os.path.expanduser('~/.bashrc')
-        if not bindir in os.environ['PATH']:
+        if bindir not in os.environ[path_var]:
             return
-        keyword = r'^[\s](*)export PATH="{}:$PATH"'.format(bindir)
+        # keyword = r'^[\s](*)export PATH="{}:$PATH"'.format(bindir)
         # escape to handle metachars
-        pattern_prepend = f'export PATH="{bindir}:$PATH"'
-        pattern_append = f'export PATH="$PATH:{bindir}"'
-        lines = []
+        pattern_prepend = f'export {path_var}="{bindir}:${path_var}"'
+        pattern_append = f'export {path_var}="${path_var}:{bindir}"'
         with open(cfg_file) as fp:
             lines = fp.readlines()
-        for l, line in enumerate(lines):
+        for li, line in enumerate(lines):
             line = line.strip()
             if line.startswith(pattern_prepend):
-                lines[l] = line.replace(pattern_prepend, '')
-            if line.startswith(pattern_append):
-                lines[l] = line.replace(pattern_append, '')
+                lines[li] = line.replace(pattern_prepend, '')
+            elif line.startswith(pattern_append):
+                lines[li] = line.replace(pattern_append, '')
         with open(cfg_file, 'w') as fp:
             fp.writelines(lines)
-    os.environ['PATH'] = os.environ['PATH'].replace(bindir, '')
+    os.environ[path_var] = os.environ[path_var].replace(bindir, '')
 
 
 def run_cmd(cmd, cwd='.', logger=None):
     if logger:
-            logger.debug(' '.join(cmd))
+        logger.debug(' '.join(cmd))
     else:
         print(' '.join(cmd))
     proc = None
@@ -956,12 +953,11 @@ def run_cmd(cmd, cwd='.', logger=None):
         raise e
     return proc
 
+
 def run_daemon(cmd, cwd='.', logger=None):
-    proc = None
     try:
         proc = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)
-        if logger:
-            logger.debug(proc.stdout.decode(TXT_CODEC))
+        # won't be able to retrieve log from background
     except subprocess.CalledProcessError as e:
         if logger:
             logger.info(f'stdout: {e.stdout.decode(TXT_CODEC)}')
@@ -979,6 +975,16 @@ def extract_call_args(file, caller, callee):
     - only support literal args
     - will throw if an arg value is a function call itself
     """
+    def get_kwarg_value_by_type(kwarg):
+        if isinstance(kwarg.value, ast.Constant):
+            return kwarg.value.value
+        elif isinstance(kwarg.value, ast.Name):
+            return kwarg.value.id
+        elif isinstance(kwarg.value, (ast.List, ast.Tuple)):
+            return [elem.value if isinstance(elem, ast.Constant) else None for elem in kwarg.value.elts]
+        print(f'Unsupported syntax node: {kwarg.value}. Will fallback to None.')
+        return None
+
     import ast
     import importlib
     import inspect
@@ -988,75 +994,25 @@ def extract_call_args(file, caller, callee):
     sys.path.insert(0, dirname(file))
     mod = importlib.import_module(mod_name)
     parsed = ast.parse(inspect.getsource(mod))
-    raw_calls = { # lineno, args, keywords
-        'func': [],
-        'method': []
+    # lineno, args, keywords
+    caller_def = next((node for node in parsed.body if isinstance(node, ast.FunctionDef) and node.name == caller), None)
+    if not caller_def:
+        return None, None
+    func_method_calls = [def_node for def_node in caller_def.body if 'value' in dir(def_node) and isinstance(def_node.value, ast.Call)]
+    raw_calls = {
+        'func': [def_node for def_node in func_method_calls if isinstance(def_node.value.func, ast.Name) and def_node.value.func.id == callee],
+        'method': [def_node for def_node in func_method_calls if isinstance(def_node.value.func, ast.Attribute) and def_node.value.func.attr == callee]
     }
-    for node in parsed.body:
-        if not isinstance(node, ast.FunctionDef) \
-            or node.name != caller:
-            continue
-        # hit function definition
-        # next: find function calls in this function
-        caller_def = node
-        callee_calls = []
-        for node in caller_def.body:
-            if 'value' not in dir(node) or not isinstance(node.value, ast.Call):
-                continue
-            # hit a function/method call
-            call_type = None
-            found_func_call = isinstance(node.value.func, ast.Name) and node.value.func.id == callee
-            found_method_call = isinstance(node.value.func, ast.Attribute) and node.value.func.attr == callee
-            if found_func_call:
-                call_type = 'func'
-            elif found_method_call:
-                call_type = 'method'
-            else:
-                continue
-            raw_calls[call_type].append(node.value)
-
-    # collect args and kwargs
-    calls = { # lineno, args, keywords
+    # collect args and kwargs: lineno, args, keywords
+    calls = {
         'func': [],
-        'method': []
+        'method': [],
     }
-    def get_kwarg_value_const(kw):
-        return kw.value.value
-    def get_kwarg_value_name(kw):
-        return kw.value.id
-    def get_kwarg_value_list(kw):
-        nodes = kw.value.elts
-        values = []
-        for node in nodes:
-            if not isinstance(node, ast.Constant):
-                values.append(None)
-                continue
-            values.append(node.value)
-        return values
-    def get_kwarg_value_misc(kw):
-        return None
-
     for calltype, rc in raw_calls.items():
         for call in rc:
-            args = []
-            kwargs = []
-            for arg in call.args:
-                args.append(arg.value)
-            for kw in call.keywords:
-                key = kw.arg
-                if isinstance(kw.value, ast.Constant):
-                    value = get_kwarg_value_const(kw)
-                elif isinstance(kw.value, ast.Name):
-                    value = get_kwarg_value_name(kw)
-                elif isinstance(kw.value, (ast.List, ast.Tuple)):
-                    value = get_kwarg_value_list(kw)
-                else:
-                    print(f'Unsupported syntax node: {kw.value}. Will fallback to None.')
-                    value = None
-                kwargs.append((key, value))
             record = {
-                'args': args,
-                'kwargs': {k: v for k, v in kwargs},
+                'args': [arg.value for arg in call.value.args],
+                'kwargs': {kw.arg: get_kwarg_value_by_type(kw) for kw in call.value.keywords},
                 'lineno': call.lineno,
                 'end_lineno': call.end_lineno,
             }
@@ -1075,6 +1031,36 @@ def extract_class_attributes(file, classname):
     - attributes can use type-annotated assignemts (taa)
     - types of attributes without taa can be inferred from constant values
     """
+    def _get_attr_by_type(node):
+        is_type_annotated = isinstance(node, ast.AnnAssign)
+        is_assigned_with_const = isinstance(node.value, ast.Constant)
+        is_assigned_with_seq = isinstance(node.value, (ast.List, ast.Tuple))
+        is_typed_coll = is_type_annotated and 'slice' in node.annotation._fields
+        if is_typed_coll:
+            coll_type = node.annotation.value.id
+            elem_type = node.annotation.slice.id if coll_type == 'list' else node.annotation.slice.dims[0].id
+            attr_type = f'{coll_type}[{elem_type}]'
+        elif is_type_annotated:
+            attr_type = node.annotation.id
+        elif is_assigned_with_const:
+            attr_type = type(node.value.value).__name__
+        else:
+            attr_type = None
+        if not is_typed_coll and is_assigned_with_seq:
+            attr_type = 'list' if isinstance(node.value, ast.List) else 'tuple'
+        # only support constants inside list
+        # non-consts are taken as None
+        attr_value = node.value.value if is_assigned_with_const \
+            else [elem.value if isinstance(elem, ast.Constant) else None for elem in node.value.elts] if is_assigned_with_seq \
+            else None
+        if coll_type_value_mismatch := attr_type == 'tuple' and isinstance(attr_value, list):
+            attr_value = tuple(attr_value)
+        # CAUTION: must not merge with type above, but as 2nd-pass after value-fix
+        if can_infer_elem_type := not is_typed_coll and is_assigned_with_seq and len(attr_value) > 0:
+            elem_type = type(attr_value[0]).__name__
+            attr_type = f'{attr_type}[{elem_type}]'
+        return attr_type, attr_value
+
     import ast
     import importlib
     import inspect
@@ -1085,53 +1071,20 @@ def extract_class_attributes(file, classname):
     mod = importlib.import_module(mod_name)
     parsed = ast.parse(inspect.getsource(mod))
     
-    class_node = None
-    for node in parsed.body:
-        if not isinstance(node, ast.ClassDef):
-            continue
-        if node.name != classname:
-            continue
-        class_node = node
+    class_node = next((node for node in parsed.body if isinstance(node, ast.ClassDef) and node.name == classname), None)
     if not class_node:
-        return None    
-    ctor = None
-    for node in ast.walk(class_node):
-        if isinstance(node, ast.FunctionDef) and node.name == '__init__':
-            ctor = node
-            break;
+        return None
+    ctor = next((node for node in ast.walk(class_node) if isinstance(node, ast.FunctionDef) and node.name == '__init__'), None)
     if not ctor:
         return None
     # parse ctor
     names = [node.attr for node in ast.walk(ctor) if isinstance(node, ast.Attribute)]
-    # types = [node.annotation.id for node in ast.walk(ctor) if isinstance(node, ast.AnnAssign)]
-    types = []
-    values = []
-    linenos = []
-    end_linenos = []
-    for node in ast.walk(ctor):
-        if not isinstance(node, (ast.AnnAssign, ast.Assign)):
-            continue
-        attr_type = None
-        attr_value = None
-        if isinstance(node, ast.AnnAssign):
-            attr_type = node.annotation.id
-        if isinstance(node.value, ast.Constant):
-            if not attr_type:
-                attr_type = type(node.value.value).__name__
-            attr_value = node.value.value
-        elif isinstance(node.value, (ast.List, ast.Tuple)):
-            nodes = node.value.elts
-            raw_values = []
-            for nd in nodes:
-                # only support constants inside list
-                # non-consts are taken as None
-                rv = nd.value if isinstance(nd, ast.Constant) else None
-                raw_values.append(rv)
-            if not attr_type:
-                attr_type = 'list' if isinstance(nd.value, ast.List) else 'tuple'
-            attr_value = raw_values
-        types.append(attr_type)
-        values.append(attr_value)
+    assigns = [node for node in ast.walk(ctor) if isinstance(node, (ast.AnnAssign, ast.Assign))]
+    types, values, linenos, end_linenos = [], [], [], []
+    for node in assigns:
+        atype, avalue = _get_attr_by_type(node)
+        types.append(atype)
+        values.append(avalue)
         linenos.append(node.lineno)
         end_linenos.append(node.end_lineno)
     attributes = [{'name': n, 'type': t, 'default': v, 'lineno': l, 'end_lineno': e} for n, t, v, l, e in zip(names, types, values, linenos, end_linenos)]
@@ -1144,7 +1097,6 @@ def extract_local_var_assignments(file, caller, varname):
     """
     - only support regular assignments (var_name = literal_value)
     """
-    value = None
     import ast
     import importlib
     import inspect
@@ -1154,43 +1106,29 @@ def extract_local_var_assignments(file, caller, varname):
     sys.path.insert(0, dirname(file))
     mod = importlib.import_module(mod_name)
     parsed = ast.parse(inspect.getsource(mod))
-    raw_calls = { # lineno, args, keywords
-        'func': [],
-        'method': []
-    }
-    assignments = []
-    for node in parsed.body:
-        if not isinstance(node, ast.FunctionDef) \
-            or node.name != caller:
-            continue
-        # hit function definition
-        # next: find function calls in this function
-        func_def = node
-        for node in ast.walk(func_def):
-            if not isinstance(node, ast.Assign) or len(node.targets) > 1:
-                continue
-            var = node.targets[0]
-            if not isinstance(var, ast.Name) or var.id != varname:
-                continue
-            if not isinstance(node.value, ast.Constant):
-                continue
-            ass = {
-                'lineno': node.lineno,
-                'end_lineno': node.end_lineno,
-                'value': node.value.value
-            }
-            assignments.append(ass)
+
+    caller_def = next((node for node in parsed.body if isinstance(node, ast.FunctionDef) and node.name == caller), None)
+    if not caller_def:
+        return None
+    single_literal_assigns = [node for node in ast.walk(caller_def) if isinstance(node, ast.Assign) and len(node.targets) == 1 and isinstance(node.value, ast.Constant)]
+    assigns = [{
+        'lineno': node.lineno,
+        'end_lineno': node.end_lineno,
+        'value': node.value.value,
+    } for node in single_literal_assigns
+        if isinstance(node.targets[0], ast.Name) and node.targets[0].id == varname
+    ]
     if mod_name in sys.modules:
         sys.modules.pop(mod_name)
-    return assignments
+    return assigns
 
 
 def extract_imported_modules(file):
-    def _extract_import_module(node):
-        return [alias.name for alias in node.names]
+    def _get_import_aliases(importmod):
+        return [alias.name for alias in importmod.names]
 
-    def _extract_from_module_import(node):
-        return node.module
+    def _extract_from_module_import(modnode):
+        return modnode.module
     imported = []
     import ast
     import importlib
@@ -1205,7 +1143,7 @@ def extract_imported_modules(file):
         if not isinstance(node, (ast.Import, ast.ImportFrom)):
             continue
         if isinstance(node, ast.Import):
-            extracted = _extract_import_module(node)
+            extracted = _get_import_aliases(node)
             if extracted:
                 imported += extracted
             continue
@@ -1217,25 +1155,41 @@ def extract_imported_modules(file):
     return sorted(list(set(imported)))
 
 
-def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0, withindent=True, useappend=False):
+def find_first_line_in_range(lines, keyword, linerange=(0,), algo='startswith'):
+    is_bandpass = len(linerange) > 1
+    if is_bandpass:
+        assert linerange[1] > linerange[0]
+    criteria = {
+        'startswith': lambda l, k: l.strip().startswith(k),
+        'endswith': lambda l, k: l.strip().endswith(k),
+        'contains': lambda l, k: k in l,
+    }
+    subs = lines[linerange[0]: linerange[1]] if is_bandpass else lines[linerange[0]:]
+    lineno_in_between = next((l for l, line in enumerate(subs) if criteria[algo](line, keyword)), None)
+    return lineno_in_between + linerange[0] if lineno_in_between is not None else None
+
+
+def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0, withindent=True, useappend=False, skipdups=False):
     """
-    - assume input lines all have line ends
-    - align inserted text with tags via identical indents
+    - lazy-append line-ends to input lines
+    - lazy-create list if input lines is a string (a single line)
+    - smart-indent lines according to tag indentation
     - optimize with search range slicing
     - returns original indices
     """
-    all_lines = []
+    lines = [lines] if isinstance(lines, str) else lines
+    lines = [line if line.endswith('\n') else f'{line}\n' for line in lines]
     with open(file) as fp:
         all_lines = fp.readlines()
     selected_lines = all_lines[startlineno:] if startlineno > 0 else all_lines
     # find range
     rg_insert = [None, None]
-    rg_insert[0] = next((l for l, line in enumerate(selected_lines) if line.strip().startswith(opkey) ), None)
+    rg_insert[0] = next((li for li, line in enumerate(selected_lines) if line.strip().startswith(opkey)), None)
     if rg_insert[0] is None:
         return rg_insert
-    rg_insert[1] = next((l for l, line in enumerate(selected_lines[rg_insert[0]:]) if line.strip().startswith(edkey) ), None)
+    rg_insert[1] = next((li for li, line in enumerate(selected_lines[rg_insert[0]:]) if line.strip().startswith(edkey)), None)
     if rg_insert[1] is None:
-        return (startlineno+rg_insert[0], None)
+        return startlineno+rg_insert[0], None
     # back to all lines with offset applied
     rg_insert[0] += startlineno
     rg_insert[1] += rg_insert[0]
@@ -1246,16 +1200,18 @@ def substitute_lines_between_keywords(lines, file, opkey, edkey, startlineno=0,
             indent_by_spaces += 4 if all_lines[rg_insert[0]][idt] == '\t' else 1
         assert indent_by_spaces >= 0
         lines = ['{}{}'.format(' '*indent_by_spaces, line) for line in lines]
-    if useappend:
-        rg_insert[0] = rg_insert[1]-1
-    else:
-        # remove lines in b/w
-        has_lines_between_keywords = rg_insert[1] - rg_insert[0] > 1
-        del all_lines[rg_insert[0]+1 : rg_insert[1]]
-    all_lines[rg_insert[0]+1 : rg_insert[0]+1] = lines
+    # remove duplicates
+    lines_to_insert = all_lines[rg_insert[0]+1: rg_insert[1]] + lines if useappend else lines
+    if skipdups:
+        lines_to_insert = list(dict.fromkeys(lines_to_insert))
+    # remove lines in b/w
+    has_lines_between_keywords = rg_insert[1] - rg_insert[0] > 1
+    if has_lines_between_keywords:
+        del all_lines[rg_insert[0]+1: rg_insert[1]]
+    all_lines[rg_insert[0]+1: rg_insert[0]+1] = lines_to_insert
     with open(file, 'w') as fp:
         fp.writelines(all_lines)
-    rg_inserted = [rg_insert[0], rg_insert[0]+len(lines)]
+    rg_inserted = [rg_insert[0], rg_insert[0]+len(lines_to_insert)]
     return rg_inserted
 
 
@@ -1296,26 +1252,25 @@ def zip_dir(srcdir, dstbasename=None):
     """
     zip the entire folder into a zip file under the same parent folder.
     """
-    def _zip_dir_windows(srcdir, dstbasename):
-        src_par, src_name = osp.split(srcdir)
-        if not dstbasename:
-            dstbasename = src_name
-        elif dstbasename != src_name:
-            dst_dir = osp.join(src_par, dstbasename)
-            duplicate_dir(srcdir, dst_dir)
-        out_file = osp.join(src_par, dstbasename+'.zip')
-        cmd = ['tar', '-cf', out_file, '-C', src_par, dstbasename]
+    def _zip_dir_windows(sdir, dstbn):
+        src_par, src_name = osp.split(sdir)
+        if not dstbn:
+            dstbn = src_name
+        elif dstbn != src_name:
+            dst_dir = osp.join(src_par, dstbn)
+            duplicate_dir(sdir, dst_dir)
+        out_file = osp.join(src_par, dstbn + '.zip')
+        cmd = ['tar', '-cf', out_file, '-C', src_par, dstbn]
         run_cmd(cmd, src_par)
 
-    def _zip_dir_macos(srcdir, dstbasename):
-        src_par, src_name = osp.split(srcdir)
+    def _zip_dir_macos(sdir, dstbn):
+        src_par, src_name = osp.split(sdir)
         rename_option = []
-        if not dstbasename:
-            dstbasename = src_name
-        elif dstbasename != src_name:
-            dst_dir = osp.join(src_par, dstbasename)
-            rename_option = ['-s', f'/^{src_name}/{dstbasename}/']
-        out_file = osp.join(src_par, dstbasename+'.zip')
+        if not dstbn:
+            dstbn = src_name
+        elif dstbn != src_name:
+            rename_option = ['-s', f'/^{src_name}/{dstbn}/']
+        out_file = osp.join(src_par, dstbn + '.zip')
         cmd = ['tar', '-czf', out_file, '--exclude', '.DS_Store', '--exclude', '*/__MACOSX'] + rename_option + ['-C', src_par, f'{src_name.strip(os.path.sep)}/']
         run_cmd(cmd, src_par)
 
@@ -1337,15 +1292,15 @@ def unzip_dir(srcball, destpardir):
     
 
 def duplicate_dir(srcdir, dstdir):
-    def _dup_dir_posix(srcdir, dstdir):
-        if not srcdir.endswith('/'):
-            srcdir += '/'
-        cmd = ['rsync', '-a', '--delete', srcdir, dstdir.strip('/')]
+    def _dup_dir_posix(sdir, ddir):
+        if not sdir.endswith('/'):
+            sdir += '/'
+        cmd = ['rsync', '-a', '--delete', sdir, ddir.strip('/')]
         run_cmd(cmd, '/')
 
-    def _dup_dir_windows(srcdir, dstdir):
-        cmd = ['xcopy', '/I', '/E', '/Q', '/Y', srcdir, f'{dstdir}\\']
-        run_cmd(cmd, srcdir)
+    def _dup_dir_windows(sdir, ddir):
+        cmd = ['xcopy', '/I', '/E', '/Q', '/Y', sdir, f'{ddir}\\']
+        run_cmd(cmd, sdir)
 
     if platform.system() == 'Windows':
         _dup_dir_windows(srcdir, dstdir)
@@ -1360,8 +1315,8 @@ def compare_textfiles(file1, file2, showdiff=False, contextonly=True, logger=Non
         if showdiff:
             diff_func = difflib.context_diff if contextonly else difflib.Differ().compare
             diff = diff_func(lines1, lines2)
-            lazy_logging(''.join(diff))
-    assert lines1 == lines2, f'content different: {file1} vs. {file2}'
+            lazy_logging(''.join(diff), logger)
+    return lines1 == lines2
 
 
 def lazy_logging(msg, logger=None):
@@ -1380,7 +1335,19 @@ def copy_file(src, dst, isdstdir=False):
     try:
         shutil.copy(src, dst)
     except shutil.SameFileError:
-        _logger.warning(f'source and destination are identical. will not copy: {osp.abspath(src)} -> {osp.abspath(dst)}. skipped.')
+        _logger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
+
+
+def move_file(src, dst, isdstdir=False):
+    if isdstdir:
+        par_dir = dst
+    else:
+        par_dir = osp.split(dst)[0]
+    os.makedirs(par_dir, exist_ok=True)
+    try:
+        shutil.move(src, dst)
+    except shutil.SameFileError:
+        _logger.warning(f'source and destination are identical. will SKIP: {osp.abspath(src)} -> {osp.abspath(dst)}.')
 
 
 def compare_dirs(dir1, dir2, ignoreddirpatterns=(), ignoredfilepatterns=(), showdiff=True):
@@ -1441,9 +1408,33 @@ def pack_obj(obj, topic=None, envelope=('<KK-ENV>', '</KK-ENV>'), classes=()):
     return f'{envelope[0]}{msg_str}{envelope[1]}'
 
 
+def lazy_extend_sys_path(paths):
+    sys.path = list(dict.fromkeys(sys.path+paths))
+
+
+def get_parent_dirs(file, subs=(), depth=1):
+    script_dir = osp.abspath(osp.dirname(file))
+    par_seq = osp.normpath('../'*depth)
+    root = osp.abspath(osp.join(script_dir, par_seq)) if depth > 0 else script_dir
+    return script_dir, root, *[osp.join(root, sub) for sub in subs]
+
+
+def get_child_dirs(root, subs=()):
+    return root, *[osp.join(root, sub) for sub in subs]
+
+
+def read_lines(file, striplineend=False, posix=True):
+    with open(file) as fp:
+        lines = fp.readlines()
+    if striplineend:
+        line_end = '\n' if posix else '\r\n'
+        for f, line in enumerate(lines):
+            lines[f] = line.strip(line_end)
+    return lines
+
+
 def _test():
-    ret = substitute_lines_between_keywords(['line 1\n', 'line 2\n'], '/Users/bin.luo/Desktop/tmp.txt', '<TAG', 'TAG>', useappend=True)
-    print(ret)
+    pass
 
 
 if __name__ == '__main__':
